"use strict";

import { BaseObserver, RowUpdateType } from '@powersync/common';
export class OPSQLiteConnection extends BaseObserver {
  constructor(options) {
    super();
    this.options = options;
    this.DB = options.baseDB;
    this.updateBuffer = [];
    this.DB.rollbackHook(() => {
      this.updateBuffer = [];
    });
    this.DB.updateHook(update => {
      this.addTableUpdate(update);
    });
  }
  addTableUpdate(update) {
    let opType;
    switch (update.operation) {
      case 'INSERT':
        opType = RowUpdateType.SQLITE_INSERT;
        break;
      case 'DELETE':
        opType = RowUpdateType.SQLITE_DELETE;
        break;
      case 'UPDATE':
        opType = RowUpdateType.SQLITE_UPDATE;
        break;
    }
    this.updateBuffer.push({
      table: update.table,
      opType,
      rowId: update.rowId
    });
  }
  flushUpdates() {
    if (!this.updateBuffer.length) {
      return;
    }
    const groupedUpdates = this.updateBuffer.reduce((grouping, update) => {
      const {
        table
      } = update;
      const updateGroup = grouping[table] || (grouping[table] = []);
      updateGroup.push(update);
      return grouping;
    }, {});
    const batchedUpdate = {
      groupedUpdates,
      rawUpdates: this.updateBuffer,
      tables: Object.keys(groupedUpdates)
    };
    this.updateBuffer = [];
    this.iterateListeners(l => l.tablesUpdated?.(batchedUpdate));
  }
  close() {
    return this.DB.close();
  }
  async execute(query, params) {
    const res = await this.DB.execute(query, params);
    return {
      insertId: res.insertId,
      rowsAffected: res.rowsAffected,
      rows: {
        _array: res.rows ?? [],
        length: res.rows?.length ?? 0,
        item: index => res.rows?.[index]
      }
    };
  }
  async executeRaw(query, params) {
    return await this.DB.executeRaw(query, params);
  }
  async executeBatch(query, params = []) {
    const tuple = [[query, params[0]]];
    params.slice(1).forEach(p => tuple.push([query, p]));
    const result = await this.DB.executeBatch(tuple);
    return {
      rowsAffected: result.rowsAffected ?? 0
    };
  }
  async getAll(sql, parameters) {
    const result = await this.DB.execute(sql, parameters);
    return result.rows ?? [];
  }
  async getOptional(sql, parameters) {
    const result = await this.DB.execute(sql, parameters);
    return result.rows?.[0] ?? null;
  }
  async get(sql, parameters) {
    const result = await this.getOptional(sql, parameters);
    if (!result) {
      throw new Error('Result set is empty');
    }
    return result;
  }
  async refreshSchema() {
    await this.get("PRAGMA table_info('sqlite_master')");
  }
}
//# sourceMappingURL=OPSQLiteConnection.js.map